#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include <time.h>
#include "utilities.h"
#include "Ring.h"

// -------------------------------------------------------------------------------
// -------------------------------------------------------------------------------
// The Calculation of F(X) and J*DX, as explained on pages 12-14
// -------------------------------------------------------------------------------
// -------------------------------------------------------------------------------


// -------------------------------------------------------------------------------
int Index(parameters par, int idom, int ipot, int i, int j)
{	// nst[idom] = Sum[ ns[jdom]*nt[jdom] , {jdom, 0, idom-1} ]; nst[0] = 0; see page 6
	int ns=par.ns[idom], nt=par.nt[idom], nst=par.nst[idom], i1=i, j1=j, Ns=ns-1, Nt=nt-1;
	
	if (i1 < 0)  i1 = -i1;
	if (i1 > Ns) i1 = 2*Ns-i1;
	
	if (j1 < 0)  j1 = -j1;
	if (j1 > Nt) j1 = 2*Nt-j1;
	
	return ipot+NPOT*(i1+ns*j1+nst);
}
// -------------------------------------------------------------------------------
int Index_i(parameters par, int i)
{	// see page 6
	int ns=par.ns[NDOM-1], Ns=ns-1, i1=i;
	
	if (i1 < 0)  i1 = -i1;
	if (i1 > Ns) i1 = 2*Ns-i1;
	
	return i1;
}
// -------------------------------------------------------------------------------
int Index_Ring(parameters par, int ipot, int i)
{	// see page 6
	return ipot+NPOT*Index_i(par, i);
}
// -------------------------------------------------------------------------------
void Get_Indices_From_n_v(parameters par, int n_v, int *idom, int *ipot, int *i, int *j, int *ipar)
{// see pages 11, 12
	int n_v_2D=par.n_v_2D, n1, n2, ns;

	*idom = NDOM;
	while(n_v < NPOT*par.nst[*idom]) *idom -= 1;

	if (*idom<NDOM){
		n1    = n_v - NPOT*par.nst[*idom];
		ns    = par.ns[*idom];
		*j    = n1/(NPOT*ns);
		n2    = n1 - NPOT*ns*(*j);
		*i    = n2/NPOT;
		*ipot = n2 - NPOT*(*i);
		*ipar = -1;
	}
	else
		*ipar = n_v - n_v_2D;
}
// -------------------------------------------------------------------------------
void v_To_V(parameters par, int idom, int i, int j, derivs_2D v, derivs_2D V)
{	// Writes (v) at gridpoint (i,j) in domain (idom) to the auxiliary derivs (V)
	int ipot, indx;
	
	for(ipot=0; ipot<NPOT; ipot++){
		indx = Index(par, idom,ipot,i,j);
		V.d0[ipot] = v.d0[indx];  // V
		V.d1[ipot] = v.d1[indx];  // V_s
		V.d2[ipot] = v.d2[indx];  // V_t
		V.d11[ipot]= v.d11[indx]; // V_ss
		V.d12[ipot]= v.d12[indx]; // V_st
		V.d22[ipot]= v.d22[indx]; // V_tt
	}
}
// -------------------------------------------------------------------------------
void Derivatives_st(parameters par, derivs_2D v)
{
	int idom, ipot, ns, nt, i, j, indx[NMAX];
	double p[NMAX], dp[NMAX], d2p[NMAX], q[NMAX], dq[NMAX];
	
	for(idom=0; idom<NDOM; idom++){

		ns=par.ns[idom];
		nt=par.nt[idom];

		for(ipot=0; ipot<NPOT; ipot++){

			for(j=0;j<nt;j++) {		// Calculation of Derivatives w.r.t. s-Direction
				for(i=0;i<ns;i++){	// (Chebyshev_Extremes)
					indx[i] = Index(par,idom,ipot,i,j);
					p[i]    = v.d0[indx[i]];
				}
				chebft_Extremes(p,ns,0);
				chder(0., 1., p,dp,ns);
				chder(0., 1., dp,d2p,ns);
				chebft_Extremes(dp,ns,1);
				chebft_Extremes(d2p,ns,1);
				for(i=0;i<ns;i++){
					v.d1[indx[i]]  = dp[i];
					v.d11[indx[i]] = d2p[i];
if (fabs(v.d1[indx[i]])>1.e+136){
printf("Derivatives_st: Dvd1=%e \t Dvd0=%e \t dom=%d \t pot=%d \t i=%d \t j=%d \n",v.d1[indx[i]],v.d0[indx[i]],idom,ipot,i,j);
//printf("")
}
				}
			}  

			for(i=0;i<ns;i++) {		// Calculation of Derivatives w.r.t. t-Direction
				for(j=0;j<nt;j++){	// (Chebyshev_Extremes)
					indx[j] = Index(par,idom,ipot,i,j);
					p[j]    = v.d0[indx[j]];
					q[j]    = v.d1[indx[j]];
				}
				chebft_Extremes(p,nt,0);
				chebft_Extremes(q,nt,0);
				chder(0., 1., p,dp,nt);
				chder(0., 1., dp,d2p,nt);
				chder(0., 1., q,dq,nt);
				chebft_Extremes(dp,nt,1);
				chebft_Extremes(d2p,nt,1);
				chebft_Extremes(dq,nt,1);
				for(j=0;j<nt;j++){
					v.d2[indx[j]]  = dp[j];
					v.d22[indx[j]] = d2p[j];
					v.d12[indx[j]] = dq[j];
				}
			}

		}
	}

}
// -------------------------------------------------------------------------------
void Derivatives_s(parameters par, derivs_1D w)
{
	int ns=par.ns[NDOM-1], ipot, i, indx_R[NMAX];
	double p[NMAX], dp[NMAX], d2p[NMAX];
	
	for(ipot=0; ipot<NPOT; ipot++){
		for(i=0;i<ns;i++){	// Calculation of Derivatives w.r.t. s-Direction (Chebyshev_Extremes)
			indx_R[i] = Index_Ring(par,ipot,i);
			p[i]      = w.d0[indx_R[i]];
		}
		chebft_Extremes(p,ns,0);
		chder(0., 1., p,dp,ns);
		chder(0., 1., dp,d2p,ns);
		chebft_Extremes(dp,ns,1);
		chebft_Extremes(d2p,ns,1);
		for(i=0;i<ns;i++){
			w.d1[indx_R[i]]  = dp[i];
			w.d11[indx_R[i]] = d2p[i];
		}
	}  
}
//  -----------------------------------------------
void Get_vwGpar_From_X(parameters par, double *X, derivs_2D v, derivs_1D w, derivs_1D G, parameters *parX)
{
	// Potentials are read in from X and stored appropriately.
	// In all but domain 3 (the ring), 
	// the value of the function is stored directly in v[n_v]. In the
	// ring, x_ij = w_i + y,t*v_ij^(4;ipot), v_i0^(4;ipot)=0, 
	// where x is either the potential omega,B or mu. Instead of u
	// (pot 0), the surface function G is read in.
	int n_2D=par.n_2D, n, n_v, idom, ipot, ipar, i, j, ns, Ns;
	
	for(n=0; n<n_2D; n++){
		n_v  = par.n_v_of_n[n];
		Get_Indices_From_n_v(par, n_v, &idom, &ipot, &i, &j, &ipar);
		
		if (idom < NDOM-1 || (idom == NDOM-1 && j > 0))     v.d0[n_v]   = X[n];
		else{
			v.d0[n_v]   = 0.;
			if(ipot > 0) w.d0[Index_Ring(par,ipot,i)] = X[n];
			else		 G.d0[i]                      = X[n];
		}
		if(idom == 0 && i == par.ns[0]-1) v.d0[Index(par, 1, ipot, par.ns[1]-1, j)] = X[n];
//		if(idom == 1 && i == 0)           v.d0[Index(par, 2, ipot,           0, j)] = X[n];
	}
	Derivatives_st(par, v);

	ns = par.ns[NDOM-1]; Ns=ns-1;
	chebft_Extremes(G.d0, ns,   0);
	chder(0., 1., G.d0, G.d1,  ns);
	chder(0., 1., G.d1, G.d11, ns);
	chebft_Extremes(G.d0,  ns, 1);
	chebft_Extremes(G.d1,  ns, 1);
	chebft_Extremes(G.d11, ns, 1);

	(*parX).xi  = G.d0[0];
	(*parX).xo  = G.d0[Ns];
//	(*parX).x0 = 4.0*(*parX).xo;
//	(*parX).x1 = 0.7*(*parX).xi;

	(*parX).V0  = X[n_2D + 0]; // X[n_2D + 0] = V0
	(*parX).Om  = X[n_2D + 1]; // X[n_2D + 1] = OmR = Omega_Ring
//	(*parX).eps = X[n_2D + 2]; // X[n_2D + 2] = eps
}
// -------------------------------------------------------------------------------
void F_of_X(parameters par, double *X, derivs_2D v, derivs_1D w, derivs_1D G, parameters *parX, double *F)
{
	int ntotal=par.ntotal, n_v_total=par.n_v_total, 
		idom, i, j, ns, Ns, nt, Nt, n, n_v;
	double *FF;
	derivs_2D V;

	allocate_derivs_2D(&V, NPOT);
	FF = dvector(0,n_v_total-1);
	fill0_dvector(FF, 0, n_v_total);

	*parX = par;
	Get_vwGpar_From_X(par, X, v, w, G, parX);
	NonLinRingBoundary(*parX, w, G);
	Derivatives_s(par, w);

	for(idom=0; idom<NDOM; idom++){
		ns=par.ns[idom];		Ns=ns-1;
		nt=par.nt[idom];		Nt=nt-1;
		for(j=0;j<nt;j++) {
			for(i=0;i<ns;i++){
				if (i*j*(Ns-i)*(Nt-j) != 0)
					NonLinFieldEqns(*parX, idom, i, j, v, w, G, V, FF);
				else
					NonLinBoundEqns(*parX, idom, i, j, v, w, G, V, FF);
			}
		}
	}
	NonLinParamEqns(par, *parX, v, w, G, V, FF);
	for(n=0; n<ntotal; n++){
		n_v  = par.n_v_of_n[n];
		F[n] = FF[n_v];
	}
	
	free_derivs_2D(&V, NPOT);
	free_dvector(FF, 0, n_v_total-1);
}
// -------------------------------------------------------------------------------
void J_times_DX(parameters par, double *X, double *DX, derivs_2D v, derivs_1D w, derivs_1D G,
        parameters parX, double *JDX)
{
	int ntotal=par.ntotal, n_v_total=par.n_v_total,
		idom, i, j, ns, Ns, nt, Nt, n, n_v;
	parameters DparX=par;
//	DparX.x0 = X0FACT*DparX.xo;
//	DparX.x1 = X1FACT*DparX.xi;
	double *JJDX;
	derivs_1D DG, Dw;
	derivs_2D Dv, V, DV;
	
	
	ns = par.ns[NDOM-1];
	allocate_derivs_1D(&DG, ns);
	allocate_derivs_1D(&Dw, ns*NPOT);
	allocate_derivs_2D(&Dv, n_v_total);
	allocate_derivs_2D( &V, NPOT);
	allocate_derivs_2D(&DV, NPOT);

	JJDX = dvector(0,n_v_total-1);
	fill0_dvector(JJDX, 0, n_v_total);

	Get_vwGpar_From_X(par, DX, Dv, Dw, DG, &DparX);
	LinRingBoundary(parX, DparX, w, G, Dw, DG);
	Derivatives_s(par, Dw);
	for(idom=0; idom<NDOM; idom++){
		ns=par.ns[idom];		Ns=ns-1;
		nt=par.nt[idom];		Nt=nt-1;
		for(j=0;j<nt;j++) {
			for(i=0;i<ns;i++){
				ns=par.ns[idom]; Ns=ns-1;
				nt=par.nt[idom]; Nt=nt-1;
				if (i*j*(Ns-i)*(Nt-j) != 0)
					LinFieldEqns(parX, DparX, idom, i, j, v, w, G, V, Dv, Dw, DG, DV, JJDX);
				else
					LinBoundEqns(parX, DparX, idom, i, j, v, w, G, V, Dv, Dw, DG, DV, JJDX);
			}
		}
	}
	
	LinParamEqns(par, parX, DparX, v, w, G, V, Dv, Dw, DG, DV, JJDX);
	for(n=0; n<ntotal; n++){
		n_v    = par.n_v_of_n[n];
		JDX[n] = JJDX[n_v];
	}
	
	ns = par.ns[NDOM-1];
	free_derivs_1D(&DG, ns);
	free_derivs_1D(&Dw, ns*NPOT);
	free_derivs_2D(&Dv, n_v_total);
	free_derivs_2D( &V, NPOT);
	free_derivs_2D(&DV, NPOT);

	free_dvector(JJDX, 0, n_v_total-1);
}
