#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include <time.h>
#include "utilities.h"
#include "Ring.h"

//-----------------------------------------------------------------------------
// 2nd and 4th order finite differencing
// -------------------------------------------------------------------------------
// -------------------------------------------------------------------------------
void GetFD_Dv_a(parameters par, derivs_2D Dv, int idom, int ipot, int i, int j)
{	// Derivatives of (Dv) with respect to (a):
	int
		i00 = Index(par, idom, ipot, i  , j), ns  = par.ns[idom], Ns = ns-1, 
		ip1 = Index(par, idom, ipot, i+1, j), ip2 = Index(par, idom, ipot, i+2, j),
		im1 = Index(par, idom, ipot, i-1, j), im2 = Index(par, idom, ipot, i-2, j);
	double ha = Pih/Ns, ga = 1./ha, ga2 = ga*ga;	// ha: Stepsize with respect to (a)

	if (i*(Ns-i) != 0){ // here: Dv.d1  = Dv_a,  Dv.d11 = Dv_(aa)
		// Dv.d1 = Dv_a, Dv.d11 = Dv_(aa):
		if (par.PreCond_FD_Order == 2){
			Dv.d1[i00]  = 0.5*ga*(Dv.d0[ip1]-Dv.d0[im1]);
			Dv.d11[i00] = ga2*(Dv.d0[ip1]+Dv.d0[im1]-2.*Dv.d0[i00]);
		}
		if (par.PreCond_FD_Order == 4){
			Dv.d1[i00]  = ga*(                   2.*(Dv.d0[ip1]-Dv.d0[im1])/3.
									- (Dv.d0[ip2]-Dv.d0[im2])/12.);
			Dv.d11[i00] = ga2*(-2.5*Dv.d0[i00] + 4.*(Dv.d0[ip1]+Dv.d0[im1])/3.
									- (Dv.d0[ip2]+Dv.d0[im2])/12.);
		}
	}
	else{ // here: Dv.d1  = Dv_(aa),  Dv.d11 = Dv_(aaaa)
		double ga4  = ga2*ga2;
		if (par.PreCond_FD_Order == 2){
			Dv.d1[i00]  = 2.*ga2*(Dv.d0[ip1]-Dv.d0[i00]); 
			Dv.d11[i00] = 2.*ga4*(3.*Dv.d0[i00]-4.*Dv.d0[ip1]+Dv.d0[ip2]); 
		}
		if (par.PreCond_FD_Order == 4){
			Dv.d1[i00]  = ga2*(-2.5*Dv.d0[i00] + 4.*(Dv.d0[ip1]+Dv.d0[im1])/3.
								- (Dv.d0[ip2]+Dv.d0[im2])/12.);
			Dv.d11[i00] = ga4*(  6.*Dv.d0[i00] - 4.*(Dv.d0[ip1]+Dv.d0[im1])
								+ (Dv.d0[ip2]+Dv.d0[im2]));
		}
	}
	
if (fabs(Dv.d1[i00])>1.e+6){
printf("GetFD_Dv_a: Dvd1=%e \t dom=%d \t pot=%d \t i=%d \t j=%d \n",Dv.d1[i00],idom,ipot,i,j);
//printf("")
}


}
// -------------------------------------------------------------------------------
void GetFD_Dv_b(parameters par, derivs_2D Dv, int idom, int ipot, int i, int j)
{	// Derivatives of (Dv) with respect to (b):
	int
		i00 = Index(par, idom, ipot, i, j),   nt  = par.nt[idom], Nt = nt-1, 
		ip1 = Index(par, idom, ipot, i, j+1), ip2 = Index(par, idom, ipot, i, j+2),
		im1 = Index(par, idom, ipot, i, j-1), im2 = Index(par, idom, ipot, i, j-2);
	double hb = Pih/Nt, gb = 1./hb, gb2 = gb*gb;	// hb: Stepsize with respect to (b)

	if (j*(Nt-j) != 0){ // here: Dv.d2  = Dv_b,  Dv.d22 = Dv_(bb)
		// Dv.d2      = Dv_b, Dv.d22 = Dv_(bb):
		if (par.PreCond_FD_Order == 2){
			Dv.d2[i00]  = 0.5*gb*(Dv.d0[ip1]-Dv.d0[im1]);
			Dv.d22[i00] = gb2*(Dv.d0[ip1]+Dv.d0[im1]-2.*Dv.d0[i00]);
		}
		if (par.PreCond_FD_Order == 4){
			Dv.d2[i00]  = gb*(                   2.*(Dv.d0[ip1]-Dv.d0[im1])/3. -
									(Dv.d0[ip2]-Dv.d0[im2])/12.);
			Dv.d22[i00] = gb2*(-2.5*Dv.d0[i00] + 4.*(Dv.d0[ip1]+Dv.d0[im1])/3. -
									(Dv.d0[ip2]+Dv.d0[im2])/12.);
		}
	}
	else{ // here: Dv.d2  = Dv_(bb),  Dv.d22 = Dv_(bbbb)
		double gb4  = gb2*gb2;
		// Dv.d2  = Dv_bb, Dv.d22 = Dv_(bbbb):
		if (par.PreCond_FD_Order == 2){
			Dv.d2[i00]  = 2.*gb2*(Dv.d0[ip1]-Dv.d0[i00]); 
			Dv.d22[i00] = 2.*gb4*(3.*Dv.d0[i00]-4.*Dv.d0[ip1]+Dv.d0[ip2]); 
		}
		if (par.PreCond_FD_Order == 4){
			Dv.d2[i00]  = gb2*(-2.5*Dv.d0[i00] + 4.*(Dv.d0[ip1]+Dv.d0[im1])/3.
							- (Dv.d0[ip2]+Dv.d0[im2])/12.);
			Dv.d22[i00] = gb4*(  6.*Dv.d0[i00] - 4.*(Dv.d0[ip1]+Dv.d0[im1])
							+ (Dv.d0[ip2]+Dv.d0[im2]));
		}
	}
}
// -------------------------------------------------------------------------------
void GetFD_Dv_ab_FD2(parameters par, derivs_2D Dv, int idom, int ipot, int i, int j)
{	// Derivatives of (Dv) with respect to (ab), Finite Difference Order 2:
	int ns=par.ns[idom], Ns=ns-1, nt=par.nt[idom], Nt=nt-1,
		icc = Index(par,idom,ipot, i  , j  ),
		ipc = Index(par,idom,ipot, i+1, j  ),
		icp = Index(par,idom,ipot, i  , j+1),
		imc = Index(par,idom,ipot, i-1, j  ),
		icm = Index(par,idom,ipot, i  , j-1),
		ipp = Index(par,idom,ipot, i+1, j+1),
		ipm = Index(par,idom,ipot, i+1, j-1),
		imp = Index(par,idom,ipot, i-1, j+1),
		imm = Index(par,idom,ipot, i-1, j-1);
	double // ha: Stepsize with respect to (a), hb: Stepsize with respect to (b)
		ha  = Pih/Ns, ga = 1./ha, ga2 = ga*ga, 
		hb  = Pih/Nt, gb = 1./hb, gb2 = gb*gb;

	if (i*(Ns-i)*j*(Nt-j) != 0) // here: Dv.d12 = Dv_(ab)
		Dv.d12[icc]  = 0.25*ga*gb*(Dv.d0[ipp]+Dv.d0[imm]-Dv.d0[ipm]-Dv.d0[imp]);
	else{
		if (i*(Ns-i) != 0)  // here: j*(Nt-j) = 0: Dv.d12 = Dv_(abb)    
			Dv.d12[icc]  = ga*gb2*((Dv.d0[ipp]-Dv.d0[ipc])-(Dv.d0[imp]-Dv.d0[imc]));
		else{
			if (j*(Nt-j) != 0)  // here: i*(Nt-i) = 0: Dv.d12 = Dv_(aab)
				Dv.d12[icc]  = ga2*gb*((Dv.d0[ipp]-Dv.d0[icp])-(Dv.d0[ipm]-Dv.d0[icm]));
			else                // here: i*(Ns-i)*j*(Nt-j) == 0: Dv.d12 = Dv_(aabb)
				Dv.d12[icc]  = 4.*ga2*gb2*(Dv.d0[ipp]+Dv.d0[icc]-Dv.d0[ipc]-Dv.d0[icp]);
		}
	}
}
// -------------------------------------------------------------------------------
void GetFD_Dv_ab_FD4(parameters par, derivs_2D Dv, int idom, int ipot, int i, int j)
{	// Derivatives of (Dv) with respect to (ab), Finite Difference Order 4:
	int ns  = par.ns[idom], Ns = ns-1, nt  = par.nt[idom], Nt = nt-1,

		im2m2 = Index(par, idom, ipot, i-2, j-2), im1m2 = Index(par, idom, ipot, i-1, j-2),
		i00m2 = Index(par, idom, ipot, i,   j-2), 
		ip1m2 = Index(par, idom, ipot, i+1, j-2), ip2m2 = Index(par, idom, ipot, i+2, j-2),

		im2m1 = Index(par, idom, ipot, i-2, j-1), im1m1 = Index(par, idom, ipot, i-1, j-1),
		i00m1 = Index(par, idom, ipot, i,   j-1), 
		ip1m1 = Index(par, idom, ipot, i+1, j-1), ip2m1 = Index(par, idom, ipot, i+2, j-1),

		im200 = Index(par, idom, ipot, i-2, j  ), im100 = Index(par, idom, ipot, i-1, j  ),
		i0000 = Index(par, idom, ipot, i,   j  ), 
		ip100 = Index(par, idom, ipot, i+1, j  ), ip200 = Index(par, idom, ipot, i+2, j  ),

		im2p1 = Index(par, idom, ipot, i-2, j+1), im1p1 = Index(par, idom, ipot, i-1, j+1),
		i00p1 = Index(par, idom, ipot, i,   j+1), 
		ip1p1 = Index(par, idom, ipot, i+1, j+1), ip2p1 = Index(par, idom, ipot, i+2, j+1),

		im2p2 = Index(par, idom, ipot, i-2, j+2), im1p2 = Index(par, idom, ipot, i-1, j+2),
		i00p2 = Index(par, idom, ipot, i,   j+2), 
		ip1p2 = Index(par, idom, ipot, i+1, j+2), ip2p2 = Index(par, idom, ipot, i+2, j+2);
	double // ha: Stepsize with respect to (a), hb: Stepsize with respect to (b)
		ha  = Pih/Ns, ga = 1./ha, ga2 = ga*ga, 
		hb  = Pih/Nt, gb = 1./hb, gb2 = gb*gb;

	if (i*(Ns-i) != 0){
		double // a(jj) = [Dv_a at Position (i,jj) with jj \in {j-2, j-1, j, j+1, j+2}]
			am2 = ga*(2.*(Dv.d0[ip1m2]-Dv.d0[im1m2])/3. - (Dv.d0[ip2m2]-Dv.d0[im2m2])/12.),
			am1 = ga*(2.*(Dv.d0[ip1m1]-Dv.d0[im1m1])/3. - (Dv.d0[ip2m1]-Dv.d0[im2m1])/12.),
			a00 = ga*(2.*(Dv.d0[ip100]-Dv.d0[im100])/3. - (Dv.d0[ip200]-Dv.d0[im200])/12.),
			ap1 = ga*(2.*(Dv.d0[ip1p1]-Dv.d0[im1p1])/3. - (Dv.d0[ip2p1]-Dv.d0[im2p1])/12.),
			ap2 = ga*(2.*(Dv.d0[ip1p2]-Dv.d0[im1p2])/3. - (Dv.d0[ip2p2]-Dv.d0[im2p2])/12.);
		if (j*(Nt-j) != 0) 	// here: i*(Ns-i)*j*(Nt-j) != 0; Dv.d12 = Dv_(ab)
			Dv.d12[i0000] = gb*(2.*(ap1-am1)/3. - (ap2-am2)/12.);
		else				// here: i*(Ns-i) != 0, j*(Nt-j) = 0: Dv.d12 = Dv_(abb)   
			Dv.d12[i0000] = gb2*(-2.5*a00 + 4.*(ap1+am1)/3. - (ap2+am2)/12.);
	}
	else{
		double  // a(jj) = [Dv_(aa) at Position (i,jj) with jj \in {j-2, j-1, j, j+1, j+2}]
			am2 = ga2*(-2.5*Dv.d0[i00m2] + 4.*(Dv.d0[ip1m2]+Dv.d0[im1m2])/3.
						- (Dv.d0[ip2m2]+Dv.d0[im2m2])/12.),
			am1 = ga2*(-2.5*Dv.d0[i00m1] + 4.*(Dv.d0[ip1m1]+Dv.d0[im1m1])/3.
						- (Dv.d0[ip2m1]+Dv.d0[im2m1])/12.),
			a00 = ga2*(-2.5*Dv.d0[i0000] + 4.*(Dv.d0[ip100]+Dv.d0[im100])/3.
						- (Dv.d0[ip200]+Dv.d0[im200])/12.),
			ap1 = ga2*(-2.5*Dv.d0[i00p1] + 4.*(Dv.d0[ip1p1]+Dv.d0[im1p1])/3.
						- (Dv.d0[ip2p1]+Dv.d0[im2p1])/12.),
			ap2 = ga2*(-2.5*Dv.d0[i00p2] + 4.*(Dv.d0[ip1p2]+Dv.d0[im1p2])/3.
						- (Dv.d0[ip2p2]+Dv.d0[im2p2])/12.);
		if (j*(Nt-j) != 0) 	// here: i*(Ns-i) = 0, j*(Nt-j) != 0; Dv.d12 = Dv_(aab)
			Dv.d12[i0000] = gb*(2.*(ap1-am1)/3. - (ap2-am2)/12.);
		else				// here: i*(Ns-i) = 0, j*(Nt-j) = 0: Dv.d12 = Dv_(aabb)   
			Dv.d12[i0000] = gb2*(-2.5*a00 + 4.*(ap1+am1)/3. - (ap2+am2)/12.);
	}
}
// -------------------------------------------------------------------------------
void GetFD_Dv_st(parameters par, derivs_2D Dv, int idom, int ipot, int i, int j)
{	// Derivatives of (Dv) with respect to (s,t):
	int ns=par.ns[idom], Ns=ns-1, nt=par.nt[idom], Nt=nt-1, 
		indx=Index(par, idom,ipot, i, j);
	double 
		a=Pih*i/Ns, sa=sin(a), ca=cos(a), s2a=2.*sa*ca, s=sa*sa, c2a=1.-2.*s,
		b=Pih*j/Nt, sb=sin(b), cb=cos(b), s2b=2.*sb*cb, t=sb*sb, c2b=1.-2.*t;

	GetFD_Dv_a (par, Dv, idom, ipot, i, j);
	GetFD_Dv_b (par, Dv, idom, ipot, i, j);
	if (par.PreCond_FD_Order == 2) GetFD_Dv_ab_FD2(par, Dv, idom, ipot, i, j);
	if (par.PreCond_FD_Order == 4) GetFD_Dv_ab_FD4(par, Dv, idom, ipot, i, j);

	if (i*(Ns-i) != 0){
		Dv.d1[indx]  = Dv.d1[indx]/s2a; // Dv_s = Dv_a/s2a
		// Dv_(ss) = (Dv_(aa) - 2*c2a*Dv_s)/(s2a^2)
		Dv.d11[indx] = (Dv.d11[indx] - 2.*c2a*Dv.d1[indx])/(4.*s*(1.-s)); 
	}
	else{
		Dv.d1[indx]  = 0.5*c2a*Dv.d1[indx]; // lim_{s2a->0}[Dv_s] = c2a*Dv_(aa)/2
		// lim_{s2a->0}[Dv_(ss)] = [Dv_(aaaa)+8*c2a*Dv_s]/12
		Dv.d11[indx] = (Dv.d11[indx]+8.*c2a*Dv.d1[indx])/12.;
	}

	if (j*(Nt-j) != 0){
		Dv.d2[indx]  = Dv.d2[indx]/s2b; // Dv_t = Dv_b/s2b
		// Dv_(tt) = (Dv_(bb) - 2*c2b*Dv_t)/(s2b^2)
		Dv.d22[indx] = (Dv.d22[indx] - 2.*c2b*Dv.d2[indx])/(4.*t*(1.-t));
	}
	else{
		Dv.d2[indx]  = 0.5*c2b*Dv.d2[indx]; // lim_{s2b->0}[Dv_t] = c2b*Dv_(bb)/2
		// lim_{s2b->0}(Dv_tt) = (Dv_(bbbb)+8*c2b*Dv_t)/12
		Dv.d22[indx] = (Dv.d22[indx]+8.*c2b*Dv.d2[indx])/12.;
	}

	if (i*(Ns-i)*j*(Nt-j) != 0) 
		Dv.d12[indx] =  Dv.d12[indx]/(s2a*s2b);	// Dv_(st) = Dv_(ab)/(s2a*s2b)
	else{
		if (i*(Ns-i) != 0)  // lim_{s2b->0}(Dv_(st)) = c2b*Dv_(abb)/(2*s2a)
			Dv.d12[indx]  = 0.5*c2b*Dv.d12[indx]/s2a;
		else{
			if (j*(Nt-j) != 0) // lim_{s2a->0}(Dv_(st)) = c2a*Dv_(aab)/(2*s2b)
				Dv.d12[indx]  = 0.5*c2a*Dv.d12[indx]/s2b;
			else  // lim_{s2a->0, s2b->0}(Dv_(st)) = c2a*c2b*Dv_(aabb)/4
				Dv.d12[indx]  = 0.25*c2a*c2b*Dv.d12[indx];
		}
	}
	
	
if (fabs(Dv.d1[indx])>1.e+6){
printf("GetFD_Dv_st: Dvd1=%e \t dom=%d \t pot=%d \t i=%d \t j=%d \n",Dv.d1[indx],idom,ipot,i,j);
//printf("")
}
}
// -------------------------------------------------------------------------------
void GetFD_DG_s(parameters par, derivs_1D DG, int i)
{	// Derivatives of (DG) with respect to (s):
	int idom=NDOM-1, ns=par.ns[idom], Ns=ns-1, 
		ip1 = Index_i(par, i+1), ip2 = Index_i(par, i+2), 
		im1 = Index_i(par, i-1), im2 = Index_i(par, i-2); 
	double ha  = Pih/Ns, ga = 1./ha, ga2 = ga*ga, // ha: Stepsize with respect to (a)
		a=ha*i, sa=sin(a), ca=cos(a), s2a=2.*sa*ca, s=sa*sa, c2a=1.-2.*s;

	if (i*(Ns-i) != 0){ 
		// DG.d1    = DG_a, DG.d11 = DG_(aa):
		if (par.PreCond_FD_Order == 2){
			DG.d1[i]  = 0.5*ga*(DG.d0[ip1]-DG.d0[im1]);
			DG.d11[i] = ga2*(DG.d0[ip1]+DG.d0[im1]-2.*DG.d0[i]);
		}
		if (par.PreCond_FD_Order == 4){
			DG.d1[i]  = ga*(                 2.*(DG.d0[ip1]-DG.d0[im1])/3. - (DG.d0[ip2]-DG.d0[im2])/12.);
			DG.d11[i] = ga2*(-2.5*DG.d0[i] + 4.*(DG.d0[ip1]+DG.d0[im1])/3. - (DG.d0[ip2]+DG.d0[im2])/12.);
		}
		DG.d1[i]    = DG.d1[i]/s2a; // DG_s = DG_a/s2a
		// DG_(ss)  = (DG_(aa) - 2*c2a*DG_s)/(s2a^2)
		DG.d11[i]   = (DG.d11[i] - 2.*c2a*DG.d1[i])/(4.*s*(1.-s)); 
	}
	else{ 
		double ga4  = ga2*ga2;
		// DG.d1    = DG_aa, DG.d11 = DG_(aaaa):
		if (par.PreCond_FD_Order == 2){
			DG.d1[i]  = 2.*ga2*(DG.d0[ip1]-DG.d0[i]); 
			DG.d11[i] = 2.*ga4*(3.*DG.d0[i]-4.*DG.d0[ip1]+DG.d0[ip2]); 
		}
		if (par.PreCond_FD_Order == 4){
			DG.d1[i]  = ga2*(-2.5*DG.d0[i] + 4.*(DG.d0[ip1]+DG.d0[im1])/3. - (DG.d0[ip2]+DG.d0[im2])/12.);
			DG.d11[i] = ga4*(  6.*DG.d0[i] - 4.*(DG.d0[ip1]+DG.d0[im1])    + (DG.d0[ip2]+DG.d0[im2]));
		}
		DG.d1[i]    = 0.5*c2a*DG.d1[i]; // lim_{s2a->0}[DG_s] = c2a*DG_(aa)/2
		// lim_{s2a->0}[DG_(ss)] = [DG_(aaaa)+8*c2a*DG_s]/12
		DG.d11[i]   = (DG.d11[i]+8.*c2a*DG.d1[i])/12.;
	}
}
// -------------------------------------------------------------------------------
void GetFD_Dw_s(parameters par, derivs_1D Dw, int ipot, int i)
{	// Derivatives of (Dw) with respect to (s):
	int idom=NDOM-1, ns=par.ns[idom], Ns=ns-1,
		i00 = Index_Ring(par, ipot, i),
		ip1 = Index_Ring(par, ipot, i+1),
		ip2 = Index_Ring(par, ipot, i+2),
		im1 = Index_Ring(par, ipot, i-1),
		im2 = Index_Ring(par, ipot, i-2);
	double ha  = Pih/Ns, ga = 1./ha, ga2 = ga*ga, // ha: Stepsize with respect to (a)
		a=ha*i, sa=sin(a), ca=cos(a), s2a=2.*sa*ca, s=sa*sa, c2a=1.-2.*s;

	if (i*(Ns-i) != 0){ 
		// Dw.d1      = Dw_a, Dw.d11 = Dw_(aa):
		if (par.PreCond_FD_Order == 2){
			Dw.d1[i00]  = 0.5*ga*(Dw.d0[ip1]-Dw.d0[im1]);
			Dw.d11[i00] = ga2*(Dw.d0[ip1]+Dw.d0[im1]-2.*Dw.d0[i00]);
		}
		if (par.PreCond_FD_Order == 4){
			Dw.d1[i00]  = ga*(                   2.*(Dw.d0[ip1]-Dw.d0[im1])/3.
				- (Dw.d0[ip2]-Dw.d0[im2])/12.);
			Dw.d11[i00] = ga2*(-2.5*Dw.d0[i00] + 4.*(Dw.d0[ip1]+Dw.d0[im1])/3.
				- (Dw.d0[ip2]+Dw.d0[im2])/12.);
		}
		Dw.d1[i00] = Dw.d1[i00]/s2a; // Dw_s = Dw_a/s2a
		// Dw_(ss)  = (Dw_(aa) - 2*c2a*Dw_s)/(s2a^2)
		Dw.d11[i00]   = (Dw.d11[i00] - 2.*c2a*Dw.d1[i00])/(4.*s*(1.-s)); 
	}
	else{ 
		double ga4  = ga2*ga2;
		// Dw.d1      = Dw_aa, Dw.d11 = Dw_(aaaa):
		if (par.PreCond_FD_Order == 2){
			Dw.d1[i00]  = 2.*ga2*(Dw.d0[ip1]-Dw.d0[i00]); 
			Dw.d11[i00] = 2.*ga4*(3.*Dw.d0[i00]-4.*Dw.d0[ip1]+Dw.d0[ip2]); 
		}
		if (par.PreCond_FD_Order == 4){
			Dw.d1[i00]  = ga2*(-2.5*Dw.d0[i00] + 4.*(Dw.d0[ip1]+Dw.d0[im1])/3.
				- (Dw.d0[ip2]+Dw.d0[im2])/12.);
			Dw.d11[i00] = ga4*(  6.*Dw.d0[i00] - 4.*(Dw.d0[ip1]+Dw.d0[im1])
				+ (Dw.d0[ip2]+Dw.d0[im2]));
		}
		Dw.d1[i00]    = 0.5*c2a*Dw.d1[i00]; // lim1_{s2a->0}[Dw_s] = c2a*Dw_(aa)/2
		// lim1_{s2a->0}[Dw_(ss)] = [Dw_(aaaa)+8*c2a*Dw_s]/12
		Dw.d11[i00]   = (Dw.d11[i00]+8.*c2a*Dw.d1[i00])/12.;
	}
}
// -------------------------------------------------------------------------------
void GetFD_DvwGpar(parameters par, int n_v, double *DX, derivs_2D v, derivs_1D w, derivs_1D G, 
	derivs_2D Dv, derivs_1D Dw, derivs_1D DG, parameters *DparX)
{// Calculating the finite difference strcutures Dv, Dw, Dg and DparX from a unit vector DX
 // with the non-vanishing entry n_v[n]

	int n_2D=par.n_2D, ns, Ns, nt, Nt, i0, i1, ii, j0, j1, jj, idom, ipot, i, j, ipar;

	Get_Indices_From_n_v(par, n_v, &idom, &ipot, &i, &j, &ipar);

	if (idom < NDOM-1 || (idom == NDOM-1 && j > 0)) {
		ns = par.ns[idom];              nt = par.nt[idom];
		Ns = ns-1;                      Nt = nt-1;
		i0 = maximum2(0,i-2)  - i;      j0 = maximum2(0,j-2)  - j;
		i1 = minimum2(i+2,Ns) - i;      j1 = minimum2(j+2,Nt) - j;
		
		Dv.d0[n_v] = 1.;
		if(idom == 0 && i == Ns) Dv.d0[Index(par, 1, ipot, par.ns[1]-1, j)] = 1.;
//		if(idom == 1 && i ==  0) Dv.d0[Index(par, 2, ipot,           0, j)] = 1.;

		for(ii=i0; ii<=i1; ii++)
			for(jj=j0; jj<=j1; jj++){
				GetFD_Dv_st(par, Dv, idom, ipot, i+ii, j+jj);
				if(idom == 0 && i == Ns) GetFD_Dv_st(par, Dv, 1, ipot, par.ns[1]-1+ii, j+jj);
//				if(idom == 1 && i ==  0) GetFD_Dv_st(par, Dv, 2, ipot,             ii, j+jj);
			}
	}
	else{
		if(idom < NDOM){
			ns = par.ns[idom];
			Ns = ns-1;
			i0 = maximum2(0,i-2)  - i;
			i1 = minimum2(i+2,Ns) - i;
			if (ipot > 0) {
				Dw.d0[Index_Ring(par, ipot, i)] = 1.;
				for(ii=i0; ii <= i1; ii++)   GetFD_Dw_s(par, Dw, ipot, i+ii);
			}
			else{
				DG.d0[i] = 1.;
				for(ii=i0; ii <= i1; ii++)   GetFD_DG_s(par, DG, i+ii);
			}
		}
		ns = par.ns[NDOM-1]; Ns = ns-1;
		(*DparX).xi  = DG.d0[0];
		(*DparX).xo  = DG.d0[Ns];

		(*DparX).V0  = DX[n_2D + 0]; // X[n_2D + 0] = V0
		(*DparX).Om  = DX[n_2D + 1]; // X[n_2D + 1] = Om = Omega_Ring
//		(*DparX).eps = DX[n_2D + 2]; // X[n_2D + 2] = eps

		LinRingBoundary(par, *DparX, w, G, Dw, DG);
		for(ii=0;  ii  < ns; ii++) GetFD_Dw_s(par, Dw, 0, ii);
	}
}
//  -----------------------------------------------
void fill0_Dv_of_indx(int indx, derivs_2D Dv)
{
	Dv.d0[indx] = Dv.d1[indx] = Dv.d2[indx] = Dv.d11[indx] = Dv.d12[indx]= Dv.d22[indx] = 0.;
}
//  -----------------------------------------------
void fill0_DvwGpar(parameters par, int n_v, derivs_2D Dv, derivs_1D Dw, derivs_1D DG, parameters *DparX)
{
	int ns, Ns, nt, Nt,
		idom, ipot, i, j, ipar, i0, i1, ii, j0, j1, jj, indx;

	Get_Indices_From_n_v(par, n_v, &idom, &ipot, &i, &j, &ipar);

	if (idom < NDOM-1 || (idom == NDOM-1 && j > 0)) {
		ns = par.ns[idom]; Ns = ns-1;
		nt = par.nt[idom], Nt = nt-1;
 
		i0 = maximum2(0,i-2)  - i;
		i1 = minimum2(i+2,Ns) - i;
		j0 = maximum2(0,j-2)  - j;
		j1 = minimum2(j+2,Nt) - j;
		for(ii=i0; ii<=i1; ii++){
			for(jj=j0; jj<=j1; jj++){
				indx = Index(par, idom,ipot, i+ii, j+jj);
				fill0_Dv_of_indx(indx, Dv);
				if(idom == 0 && i == Ns) {
					indx = Index(par, 1, ipot, par.ns[1]-1+ii, j+jj);
					fill0_Dv_of_indx(indx, Dv);
				}
/*				if(idom == 1 && i ==  0) {
					indx = Index(par, 2, ipot, ii, j+jj);
					fill0_Dv_of_indx(indx, Dv);
				}	*/
			}
		}
	}
	else{
		int ns = par.ns[NDOM-1];
		fill0_derivs_1D(DG, ns);	
		fill0_derivs_1D(Dw, ns*NPOT);	
	}
	(*DparX).xi  = 0.;
	(*DparX).xo  = 0.;

	(*DparX).V0  = 0.;
	(*DparX).Om  = 0.;
//	(*DparX).eps = 0.;
}
// -------------------------------------------------------------------------------
void Get_boundary_i0(parameters par, int jdom, int *adom, int *ai0, int *ai1, int *aj0, int *aj1,
		int *mdom)
{
	adom[*mdom] = jdom;
	ai0[*mdom]  = 0;
	ai1[*mdom]  = 2;
	aj0[*mdom]  = aj0[0];
	aj1[*mdom]  = aj1[0];
	*mdom += 1;
}
// -------------------------------------------------------------------------------
void Get_boundary_i1(parameters par, int jdom, int *adom, int *ai0, int *ai1, int *aj0, int *aj1,
		int *mdom)
{
	adom[*mdom] = jdom;
	ai0[*mdom]  = par.ns[jdom]-3;
	ai1[*mdom]  = par.ns[jdom]-1;
	aj0[*mdom]  = aj0[0];
	aj1[*mdom]  = aj1[0];
	*mdom += 1;
}
// -------------------------------------------------------------------------------
void Get_boundary_j0(parameters par, int jdom, int *adom, int *ai0, int *ai1, int *aj0, int *aj1,
		int *mdom)
{
	adom[*mdom] = jdom;
	aj0[*mdom]  = 0;
	aj1[*mdom]  = 2;
	ai0[*mdom]  = ai0[0];
	ai1[*mdom]  = ai1[0];
	*mdom += 1;
}
// -------------------------------------------------------------------------------
void Get_boundary_j1(parameters par, int jdom, int *adom, int *ai0, int *ai1, int *aj0, int *aj1,
		int *mdom)
{
	adom[*mdom] = jdom;
	aj0[*mdom]  = par.nt[jdom]-3;
	aj1[*mdom]  = par.nt[jdom]-1;
	ai0[*mdom]  = ai0[0];
	ai1[*mdom]  = ai1[0];
	*mdom += 1;
}
// -------------------------------------------------------------------------------
void Get_gridpoints(parameters par, int idom, int ipot, int i, int j, int *mdom, int *adom, int *ai0,
		int *ai1, int *aj0, int *aj1)
{	// Getting the relevant gridpoints at which for a given unit vector DX the evaluation of the
	// field equations and boundary conditions yields a non-vanishing entry for J*DX
	// At the exit:
	//      mdom: number of domains within which relevant gridpoints are located
	//      for (jdom = 0; jdom < mdom; jdom ++) :
	//          adom[jdom]: contains the number of the domain (jdom), ranging from 0 to 4
	//           ai0[jdom]: contains the minimal s-index of gridpoints in the domain adom[jdom]
	//           ai1[jdom]: contains the maximal s-index of gridpoints in the domain adom[jdom]
	//           aj0[jdom]: contains the minimal t-index of gridpoints in the domain adom[jdom]
	//           aj1[jdom]: contains the maximal t-index of gridpoints in the domain adom[jdom]
	int jdom, ns = par.ns[idom], Ns = ns-1, nt = par.nt[idom], Nt = nt-1;

	*mdom   = 1;
	adom[0] = idom;

	ai0[0] = maximum2(0,i-2);
	ai1[0] = minimum2(i+2,Ns);
	aj0[0] = maximum2(0,j-2);
	aj1[0] = minimum2(j+2,Nt);

	if (idom == 0){
		if(ai1[0] == Ns) {
			Get_boundary_i1(par, 1, adom, ai0, ai1, aj0, aj1, mdom);
			if(aj1[0] == Nt) Get_boundary_j1(par, 2, adom, ai0, ai1, aj0, aj1, mdom);
		}
	}
	if (idom == 1){
		if(ai1[0] == Ns) Get_boundary_i1(par, 0, adom, ai0, ai1, aj0, aj1, mdom);
//		if(ai0[0] == 0)  Get_boundary_i0(par, 2, adom, ai0, ai1, aj0, aj1, mdom);
		if(aj1[0] == Nt) Get_boundary_j1(par, 2, adom, ai0, ai1, aj0, aj1, mdom);
	}
//	if (idom == 2 && ai0[0] == 0) Get_boundary_i0(par, 1, adom, ai0, ai1, aj0, aj1, mdom);
	if (idom == 2){
		if(aj0[0] == 0)  Get_boundary_j0(par, 3, adom, ai0, ai1, aj0, aj1, mdom);
		if(aj1[0] == Nt) Get_boundary_j1(par, 1, adom, ai0, ai1, aj0, aj1, mdom);
	}
	if (idom == 3 && aj0[0] == 0){
		Get_boundary_j0(par, 2, adom, ai0, ai1, aj0, aj1, mdom);
		if(j==0){
			if(ipot==0) {
				aj0[1] = 0;
				aj1[1] = par.nt[2] - 1;
				Get_boundary_j1(par, 1, adom, ai0, ai1, aj0, aj1, mdom);
			}
			aj0[0] = 0;
			aj1[0] = Nt;
		}
	}
	if (idom == NDOM || (idom == NDOM-1 && ipot == 0 && i*(Ns-i) == 0 && j == 0)){
		*mdom = NDOM;
		for(jdom=0; jdom<NDOM; jdom++){
			adom[jdom] = jdom;
			ai0[jdom]  = aj0[jdom] = 0;
			ai1[jdom]  = par.ns[jdom]-1;
			aj1[jdom]  = par.nt[jdom]-1;
		}
	}
}
// -------------------------------------------------------------------------------
void Get_JFD_Matrix(parameters par, double *X, derivs_2D v, derivs_1D w, derivs_1D G,
          parameters parX, JFD_Components *JFD)
{// Calculating JFD in the form of eqn. (16)
	int ntotal=par.ntotal, n1=par.n1, n2=par.n2, n_v_total = par.n_v_total,
		ipot,  idom,  i,  j,  ipar, ns, row, row_v, column, column_v, mcol, 
		ipot1, idom1, ii, jj, i0, i1, j0, j1, ns1, Ns1, nt1, Nt1,
		mdom,  jdom, adom[NDOM], ai0[NDOM], aj0[NDOM], ai1[NDOM], aj1[NDOM];
	parameters DparX=par;
//	DparX.x0 = X0FACTOR*Dpar.xo; // these are constants
//	DparX.x1 = X1FACTOR*Dpar.xi;

	double *DX, *JJDX;
	derivs_1D DG, Dw;
	derivs_2D Dv, V, DV;

	ns = par.ns[NDOM-1];
	allocate_derivs_1D(&DG, ns);           fill0_derivs_1D(DG, ns);
	allocate_derivs_1D(&Dw, ns*NPOT);      fill0_derivs_1D(Dw, ns*NPOT);
	allocate_derivs_2D(&Dv, n_v_total);    fill0_derivs_2D(Dv, n_v_total);
	allocate_derivs_2D( &V, NPOT);         fill0_derivs_2D(V,  NPOT);
	allocate_derivs_2D(&DV, NPOT);         fill0_derivs_2D(DV, NPOT);
	
	DX   = dvector(0, ntotal-1);           fill0_dvector(DX,   0, ntotal);
	JJDX = dvector(0, n_v_total-1);        fill0_dvector(JJDX, 0, n_v_total);

	(*JFD).m1 = (*JFD).m2 = 0;

	for(column = 0; column < ntotal; column ++){
		DX[column] = 1.;// setting the unit vector DX
		column_v   = par.n_v_of_n[column];
		Get_Indices_From_n_v(par, column_v, &idom, &ipot, &i, &j, &ipar);
		fill0_DvwGpar(par, column_v, Dv, Dw, DG, &DparX);
		GetFD_DvwGpar(par, column_v, DX, v, w, G, Dv, Dw, DG, &DparX);
		Get_gridpoints(par, idom, ipot, i, j, &mdom, adom, ai0, ai1, aj0, aj1); 
		for(jdom=0; jdom<mdom; jdom++){
			idom1 = adom[jdom];
			i0    = ai0[jdom];       i1  = ai1[jdom];
			j0    = aj0[jdom];       j1  = aj1[jdom]; 
			ns1=par.ns[idom1];       Ns1 = ns1-1;
			nt1=par.nt[idom1];       Nt1 = nt1-1;
			for(ii=i0; ii<=i1; ii++){
				for(jj=j0; jj<=j1; jj++){
					if (ii*jj*(Ns1-ii)*(Nt1-jj) != 0)
						LinFieldEqns(parX, DparX, idom1, ii, jj, v, w, G, V, Dv, Dw, DG, DV, JJDX);
					else
						LinBoundEqns(parX, DparX, idom1, ii, jj, v, w, G, V, Dv, Dw, DG, DV, JJDX);
				}
			}
		}
		LinParamEqns(par, parX, DparX, v, w, G, V, Dv, Dw, DG, DV, JJDX);
		for(jdom=0; jdom<mdom; jdom++){
			idom1 = adom[jdom]; 
			i0    = ai0[jdom]; i1 = ai1[jdom]; 
			j0    = aj0[jdom]; j1 = aj1[jdom]; 
			for(ipot1=0; ipot1<NPOT; ipot1++){
				for(ii=i0; ii<=i1; ii++){
					for(jj=j0; jj<=j1; jj++){
						row_v = Index(par,idom1, ipot1, ii, jj);
						row   = par.n_of_n_v[row_v];
						if(row >= 0 && row < n2){
							if(column<n2){ // writing columns of row No. n in J_UL
								if (fabs(JJDX[row_v]) > TINY){
									mcol                        =  (*JFD).ncols_J_UL[row];
									(*JFD).cols_J_UL[row][mcol] =  column;
									(*JFD).J_UL[row][mcol]      =  JJDX[row_v];
									(*JFD).ncols_J_UL[row]     +=  1;
									if((row < n1 && column < n1) || (row >= n1 && column >= n1)){
										if(column > row) // determining the numbers m1 and m2, see page 16
											(*JFD).m2=maximum2((*JFD).m2, column-row);
										else
											(*JFD).m1=maximum2((*JFD).m1, row-column);
									}
								}
							}
							else // writing columns of row No. n in J_UR
								(*JFD).J_UR[row][column-n2] = JJDX[row_v];
							JJDX[row_v] =  0.;
						}
					}
				}
			}
		}
		for(row=n2; row<ntotal; row++){// writing columns of row No. n in J_LL and J_LR
			row_v = par.n_v_of_n[row];
			if(column<n2) (*JFD).J_LL[row-n2][column]    = JJDX[row_v];
			else          (*JFD).J_LR[row-n2][column-n2] = JJDX[row_v];
			JJDX[row_v] = 0.;
		}
		fill0_DvwGpar(par, column_v, Dv, Dw, DG, &DparX);
		DX[column] = 0.;
	}
	ns = par.ns[NDOM-1];
	free_derivs_1D(&DG, ns);
	free_derivs_1D(&Dw, ns*NPOT);
	free_derivs_2D(&Dv, n_v_total);
	free_derivs_2D( &V, NPOT);
	free_derivs_2D(&DV, NPOT);

	free_dvector(DX,       0, ntotal-1);
	free_dvector(JJDX,     0, n_v_total-1);
}
//-----------------------------------------------------------------------------
void Get_JFD_Components(parameters par, JFD_Components *JFD)
{// Calculates the remaining components of JFD, see pages 14-18
 // Note: For the band-matrix operations we use the 'Numerical Recipes in C'
 // which provide the corresponding routines for matrices and vectors with indices
 // starting from 1.
	int ntotal = par.ntotal, n2=par.n2, n3=ntotal-n2, m1=(*JFD).m1, m2=(*JFD).m2,
		j, k, m, row_K, col_K, col_J, row_Z, col_Z, row_H, col_H, col_Z_max=0;
	double d=-1., *aux_F, *aux_X;

	aux_X = dvector(0, n2);
	aux_F = dvector(0, n2);

	for(j=0; j<n2; j++){
		row_K = j+1; // see Note
		for(k=0; k < (*JFD).ncols_J_UL[j]; k++){
			col_J = (*JFD).cols_J_UL[j][k];
			col_K = col_J+1; // see Note
			if (row_K-col_K <= m1 && col_K-row_K <= m2) // writing K_UL
				(*JFD).K_UL[row_K][m1+1+col_K-row_K] = (*JFD).J_UL[j][k];
			else{// writing Z_UL
				int n_v = par.n_v_of_n[col_J], idom, ipot, ii, jj, ipar;
				Get_Indices_From_n_v(par, n_v, &idom, &ipot, &ii, &jj, &ipar);
				if(idom <= 1)	col_Z = 1+ipot+NPOT*ii;
				else			col_Z = 1+ipot+NPOT*(ii+par.ns[1]*(par.nt[3]-jj));
				(*JFD).iZ_UL[col_Z]       = col_K;
				(*JFD).Z_UL[row_K][col_Z] = (*JFD).J_UL[j][k];
				col_Z_max = maximum2(col_Z_max, col_Z);
			}
		}
	}
	(*JFD).col_Z_max = col_Z_max;

	// Calculating the LU-decomposition of K_UL; introducing Kl_UL and iK_UL,
	// see 'Numerical Recipes in C' pages 51-54
	bandec((*JFD).K_UL, n2, m1, m2, (*JFD).Kl_UL, (*JFD).iK_UL, &d);

	// Calculating (K_UL)^(-1)*Z_UL; storing in Z_UL
	for(col_Z=1; col_Z<=col_Z_max; col_Z ++){
		for(row_K = 1; row_K <= n2; row_K++) aux_F[row_K] = (*JFD).Z_UL[row_K][col_Z];
		banbks((*JFD).K_UL, n2, m1, m2, (*JFD).Kl_UL, (*JFD).iK_UL, aux_F);
		for(row_K = 1; row_K <= n2; row_K++) (*JFD).Z_UL[row_K][col_Z] = aux_F[row_K];
	}

	// Calculating H_UL with the methods in 'Numerical Recipes in C', pages 75-77
	for(row_H=1; row_H<=col_Z_max; row_H ++){
		for(col_H=1; col_H<=col_Z_max; col_H ++){
			row_Z = (*JFD).iZ_UL[row_H];
			col_Z = col_H;
			if(row_Z>0)	(*JFD).H_UL[row_H][col_H] = (*JFD).Z_UL[row_Z][col_Z];
			else			(*JFD).H_UL[row_H][col_H] = 0.;
			if(row_H == col_H) (*JFD).H_UL[row_H][col_H]  += 1.; 
		}
	}

	// Calculating the LU-decompositionm of H_UL; introducing the index array iH_UL
	// This completes the 'UL'-Components
	ludcmp_1((*JFD).H_UL, col_Z_max, (*JFD).iH_UL, &d);

	// --- Calculating the remaining JFD-Components ---
	// Calculating (J_UL)^(-1)*J_UR; storing in J_UR
	for(k=0; k<n3; k++){
		for(j=0; j<n2; j++) aux_F[j] = (*JFD).J_UR[j][k];
		Solve_JFD_UL(par, aux_F, aux_X, (*JFD));
		for(j=0; j<n2; j++) (*JFD).K_UR[j][k] = aux_X[j];
	}

	// Calculating (K_LR)^(-1)=J_LR-J_LL*K_UR
	for(j=0; j<n3; j++){
		for(k=0; k<n3; k++){
			(*JFD).K_LR[j][k] = (*JFD).J_LR[j][k];
			for(m=0; m<n2; m++) (*JFD).K_LR[j][k] -= (*JFD).J_LL[j][m]*(*JFD).K_UR[m][k];
		}
	}

	// Calculating the LU-decomposition of (K_LR)^(-1) and storing in JFD.K_LR;
	// introducing the index array iK_LR
	ludcmp((*JFD).K_LR, n3, (*JFD).iK_LR, &d);

	free_dvector(aux_X, 0, n2);
	free_dvector(aux_F, 0, n2);
}
// -------------------------------------------------------------------------------


