#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include <time.h>
#include "utilities.h"
#include "Ring.h"

//  -----------------------------------------------
void M_and_J(parameters par, derivs_2D v, derivs_1D w, derivs_1D G, int ij, double *M, double *J)
{
// Calculates the Mass Surface Integral in domain (idom) (integration in upper half-space):

//  2) Integration over t=const=sin^2[Pih*ij/Nt] in domain 2: ['=d/ds]
//		M/2 = 1/2*Integrate[                         rho*B*[rho'(s)*nu_zeta    - zeta'(s)*nu_rho]
//		                    -omega/2*(rho*B)^3*Exp[-4*nu]* [rho'(s)*omega_zeta - zeta'(s)*omega_rho],
//		                   {t,0,1}]
//		    = 1/2*Integrate[MInt[s]/Sqrt[s*(1-s)], {s,0,1}]
//		J/2 = -1/8*Integrate[(rho*B)^3*Exp[-4*nu]*[rho'(s)*omega_zeta - zeta'(s)*omega_rho], {s,0,1}]
//		    = -1/8*Integrate[JInt[s]/Sqrt[s*(1-s)], {s,0,1}]

	int idom=NDOM-2, ns=par.ns[idom], Ns=ns-1, nt=par.nt[idom], Nt=nt-1, isignum, i, j=ij;
	double MInt[NMAX], JInt[NMAX], Laplace[4], grad_grad[16], x, y, hat_zeta,
		u, ux, uy, B, Bx, By, om, omx, omy, Bnux, Bnuy, aux_Bnu, aux_om, b=Pih*j/Nt, sb=sin(b), t=sb*sb;
	coordinate xx, yy;
	derivs_2D V;
	allocate_derivs_2D(&V, NPOT);

	for(i=0; i<ns; i++){
		double a=Pih*i/Ns, sa=sin(a), s=sa*sa;
		coord_st_To_xy(par, idom, i, j, s, t, G, &xx, &yy);
		NonLinPotentials(par, idom, i, j, s, t, xx, yy, v, w, G, V, Laplace, grad_grad);
		x   = xx.d0;    y = yy.d0;
		if(i*(Ns-i) != 0) hat_zeta = sqrt(y/(s*(1.-s))); // hat{zeta} = zeta/Sqrt[s*(1-s)]
		else              hat_zeta = sqrt(yy.d1/(1.-2.*s));
		
		u    = V.d0[0];  ux = V.d1[0];  uy = V.d2[0];
		B    = V.d0[1];  Bx = V.d1[1];  By = V.d2[1];
		om   = V.d0[2]; omx = V.d1[2]; omy = V.d2[2];
		Bnux = B*ux + Bx;
		Bnuy = B*uy + By;
		
		aux_Bnu = xx.d1*hat_zeta*s*(1.-s)*Bnuy - yy.d1/hat_zeta*x*Bnux;
		aux_om  = xx.d1*hat_zeta*s*(1.-s)*omy  - yy.d1/hat_zeta*x*omx;
		
		JInt[i] = -0.25*x*exp(-4.*u)/B*aux_om;
		MInt[i] = aux_Bnu +2.*om*JInt[i];
	}
	chebft_Extremes(MInt,ns,0);
	chebft_Extremes(JInt,ns,0);
	*M = Pih*MInt[0];
	*J = Pih*JInt[0];
	free_derivs_2D(&V, NPOT);
}
//  -----------------------------------------------
void Volume_integrals(parameters par, derivs_2D v, derivs_1D w, derivs_1D G, double *M0, double *M, double *J)
{
// Calculates Mass, Angular Momentum and two "Virial Identities" via Integrals in domain (Ndom-1)
// Integration in upper half-space, over (s,t) in [0,1]^2
	int idom, ns, Ns, nt, Nt, isignum, i, j, ii;
	double Om=par.Om, Ms_part[NMAX], Mst_part[NMAX], M0s[NMAX], M0st[NMAX],  Js[NMAX], Jst[NMAX],
				Laplace[NPOT], grad_grad[16], x, hat_rho, hat_zeta, u, B, om, mu, nu, eV, ae2V, vel, vel2,
				det, h, p, eps, epsB, aux, aux_drho_dzeta, s, t, a, b, sa, sb;
	coordinate xx, yy;
	derivs_2D V;

	allocate_derivs_2D(&V, NPOT);
	
	idom=NDOM-1; ns=par.ns[idom]; Ns=ns-1; nt=par.nt[idom]; Nt=nt-1;
	for(i=0; i<ns; i++){
		for(j=0; j<nt; j++) {
			a = Pih*i/Ns; sa=sin(a); s=sa*sa;
			b = Pih*j/Nt; sb=sin(b); t=sb*sb;
			if (i*(Ns-i)==-12){
				coord_st_To_xy(par, NDOM-2, i, 0, s, 0., G, &xx, &yy);
				NonLinPotentials(par, NDOM-2, i, 0, s, 0., xx, yy, v, w, G, V, Laplace, grad_grad);
			}
			else{
				coord_st_To_xy(par, idom, i, j, s, t, G, &xx, &yy);
				NonLinPotentials(par, idom, i, j, s, t, xx, yy, v, w, G, V, Laplace, grad_grad);
			}
			x  = xx.d0;  u = V.d0[0]; B = V.d0[1]; om = V.d0[2]; mu = V.d0[3];
			nu = u + log(B); vel = sqrt(x)*B*exp(-2.*nu)*(Om-om); vel2 = vel*vel;
			eV = exp(nu)*sqrt(1.-vel2); ae2V = -sqrt(x)*vel*B;
//			grad_nu_2 = grad_grad[GradIndex(0,0)] + grad_grad[GradIndex(1,1)]/B/B + 2./B*grad_grad[GradIndex(0,1)];
                 // = (nu,rho)^2 + (nu,zeta)^2
//			grad_om_2 = grad_grad[GradIndex(2,2)];  // = (om,rho)^2 + (om,zeta)^2
//			grad_u_mu = grad_grad[GradIndex(0,3)];

			if(j != Nt && i*(i-Ns)!=0)  hat_zeta = Sqrt(yy.d0/(s*(1.-s)*(1.-t))); // hat{zeta} = zeta/Sqrt[s*(1-s)*(1-t)]
			else{
				if (i*(i-Ns)==0 && j!=Nt) hat_zeta = Sqrt(yy.d1/((1.-2.*s)*(1.-t)));
				if (i*(i-Ns)==0 && j==Nt) hat_zeta = Sqrt(-yy.d12/((1.-2.*s)));
				if (i*(i-Ns)!=0 && j==Nt) hat_zeta = Sqrt(-yy.d2/(s*(1.-s)));
			}
			h  = par.V0 - log(eV);  // h    = V0 - V
			eos_quantities(par, h, &p, &eps, &epsB);// h(p) = Integrate[1/(eps(pp)+pp), {pp, 0, p}]
			det     = fabs(-xx.d1*yy.d2); // det = xx.d2*yy.d1-xx.d1*yy.d2, but xx.d2 =0
			aux_drho_dzeta = 0.5/hat_zeta*det/Sqrt(x);
			aux     = Pi/hat_zeta*det*exp(2.*mu)/eV*B;
			M0st[j] = aux*epsB;
			Mst_part[j]  = aux*eV*(eps+3.*p);
			Jst[j]  = -aux*ae2V/eV*(eps+p);
		}
		chebft_Extremes(M0st,nt,0); chebft_Extremes(Mst_part,nt,0); chebft_Extremes(Jst,nt,0);
		M0s[i] = Ms_part[i] = Js[i] = 0.;
		isignum=-1;
		for(j=1; j<=Nt; j++){
						int denom = (2*j-1)*(2*j+1)*isignum;
						M0s[i]  +=  M0st[j]/denom;
						Ms_part[i]  +=  Mst_part[j]/denom;
						Js[i]  +=  Jst[j]/denom;
						isignum = -isignum;
		}
		M0s[i]  =  M0st[0] - 2.* M0s[i];
		Ms_part[i]  =  Mst_part[0] - 2.* Ms_part[i];
		Js[i]  =  Jst[0] - 2.* Js[i];
	}
	chebft_Extremes(M0s, ns,0);
	chebft_Extremes(Ms_part, ns,0);
	chebft_Extremes(Js, ns,0);

	*M0        =  Pih*M0s[0];
	*J         =  Pih*Js[0];
	*M         =  Pih*Ms_part[0]+2.*Om*(*J);

	free_derivs_2D(&V, NPOT);
}

//   -----------------------------------------------
double ShedParameter(parameters par, int i, derivs_1D G)
{
	double s, t=0., beta, xm = Func_xm(par), rm = sqrt(xm), ro = sqrt(par.xo), R  =  ro-rm;
	coordinate xx, yy;

	if(i==0) s=0.;
	else     s=1.;

	coord_st_To_xy(par, NDOM-1, i, 0, s, t, G, &xx, &yy);

	if(i==0) beta = (-1.+rm/R)*yy.d1/xx.d1;
	else     beta = (-1.-rm/R)*yy.d1/xx.d1;
	
	return beta;
}
//  -----------------------------------------------
double D_ShedParameter(parameters par, parameters Dpar, int i, derivs_1D G, derivs_1D DG)
{
	double s, t=0.,
		beta,   xm = Func_xm(par),          rm = sqrt(xm),    ro = sqrt(par.xo),   R  =  ro-rm,
		Dbeta, Dxm = Func_Dxm(par, Dpar),  Drm = 0.5*Dxm/rm, Dro = 0.5*Dpar.xo/ro, DR  =  Dro - Drm;
	coordinate xx, yy, Dxx, Dyy;

	if(i==0) s=0.;
	else     s=1.;

	Dcoord_st_To_xy(par, Dpar, NDOM-1, i, 0, s, t, G, DG, &xx, &yy, &Dxx, &Dyy);

	if(i==0) {
		beta  = (-1.+rm/R)*yy.d1/xx.d1;
		Dbeta = (( Drm/R-rm*DR/(R*R))*yy.d1 + (-1.+rm/R)*Dyy.d1 - beta*Dxx.d1)/xx.d1;
	}
	else{
		beta  = (-1.-rm/R)*yy.d1/xx.d1;  
		Dbeta = ((-Drm/R+rm*DR/(R*R))*yy.d1 + (-1.-rm/R)*Dyy.d1 - beta*Dxx.d1)/xx.d1;
	}
	
	return Dbeta;
}
//  -----------------------------------------------
double h_max(parameters par, derivs_2D v, derivs_1D w, derivs_1D G)
{
// the position of the maximum value of e2V is determined in the ring for y=0
	int idom, ns, nt, i, j;
	double h_max, s, t, e2Vx, a, sa, Om=par.Om, V0=par.V0, Laplace[NPOT], grad_grad[NPOT*NPOT],
		s0, s1, x, u, B, om, ux, Bx, omx, e2V,
		cheb_u[NMAX], cheb_B[NMAX], cheb_om[NMAX], cheb_ux[NMAX], cheb_Bx[NMAX], cheb_omx[NMAX];
	coordinate xx, yy;
	derivs_2D V;

	allocate_derivs_2D(&V, NPOT);

	idom = NDOM-1;
	ns = par.ns[idom];
	nt = par.nt[idom];
	t=1.; j=nt-1;
	for(i=0; i<ns; i++){ // here I create Chebyshev coefficients
		a=Pih*i/(ns-1), sa=sin(a), s=sa*sa;
		coord_st_To_xy(par, idom, i, j, s, t, G, &xx, &yy);
		NonLinPotentials(par, idom, i, j, s, t, xx, yy, v, w, G, V, Laplace, grad_grad);
		cheb_u[i]=V.d0[0]; cheb_B[i]=V.d0[1]; cheb_om[i]=V.d0[2];
		cheb_ux[i]=V.d1[0]; cheb_Bx[i]=V.d1[1]; cheb_omx[i]=V.d1[2];
	}
	chebft_Extremes(cheb_u, ns, 0); chebft_Extremes(cheb_ux, ns, 0);
	chebft_Extremes(cheb_B, ns, 0); chebft_Extremes(cheb_Bx, ns, 0);
	chebft_Extremes(cheb_om, ns, 0); chebft_Extremes(cheb_omx, ns, 0);
	s0=0.; s1=1.0; s=0.5*(s0+s1);
	while (fabs(s0-s1)>1.e-15){
		coord_3_all_st(par, s, t, G, &xx, &yy);
		x = xx.d0;
		u = chebev(0., 1., cheb_u, ns, s); ux = chebev(0., 1., cheb_ux, ns, s);
		B = chebev(0., 1., cheb_B, ns, s); Bx = chebev(0., 1., cheb_Bx, ns, s);
		om = chebev(0., 1., cheb_om, ns, s); omx = chebev(0., 1., cheb_omx, ns, s);
		e2Vx = 2.*B*exp(2.*u)*(Bx + B*ux) - exp(-2.*u)*(Om-om)*(Om-om) + 2.*x*(Om-om)*exp(-2.*u)*(ux*(Om-om) + omx);
		if (e2Vx<0.) s0=s;
		else         s1=s;
		s = 0.5*(s0+s1);
	}
	e2V       = B*B*exp(2.*u) - x*exp(-2.*u)*(Om-om)*(Om-om);
	h_max     = par.V0 - 0.5*log(e2V);  // h    = V0 - V

	free_derivs_2D(&V, NPOT);

	return h_max;
}
//  -----------------------------------------------
double Rcirc(parameters par, derivs_2D v, derivs_1D w, derivs_1D G, int i)
{// R_circ = rho*e^(-u) is found in the equatorial plane in the ring for given i
	int j, ns, nt, idom=NDOM-1;
	double s, t, a, sa, sb, Laplace[NPOT], grad_grad[NPOT*NPOT], u;
	coordinate xx, yy;
	derivs_2D V;

	allocate_derivs_2D(&V, NPOT);

	ns = par.ns[idom];
	nt = par.nt[idom];
	a=Pih*i/(ns-1), sa=sin(a), s=sa*sa;
	j=nt-1; t=1.;

	coord_st_To_xy(par, idom, i, j, s, t, G, &xx, &yy);
	NonLinPotentials(par, idom, i, j, s, t, xx, yy, v, w, G, V, Laplace, grad_grad);
	u = V.d0[0];
	free_derivs_2D(&V, NPOT);

	return sqrt(xx.d0)*exp(-u);
}
//  -----------------------------------------------
void GetPhysicalParameters_NonLin(parameters *parX, derivs_2D v, derivs_1D w, derivs_1D G)
{	// parameters 1-6 already determined in  <Get_vwGpar_From_X>
	
	double ro = sqrt((*parX).xo), ri = sqrt((*parX).xi),
         r0 = sqrt(X0FACT*(*parX).xo), om_ss[NMAX], om_sss[NMAX];
	int i, indx;

	(*parX).r_ratio     = ri/ro;
	(*parX).beta_i      = ShedParameter(*parX, 0, G);
	(*parX).beta_o      = ShedParameter(*parX, (*parX).ns[NDOM-1]-1, G);
	(*parX).M_infty     = -r0*v.d1[Index(*parX, 0, 0, 0, 0)];
// a function omega(s,t=0),s,s will be defined so that a further derivative
// can be taken in order to calculate the angular momentum J
	for(i=0; i<(*parX).ns[0]; i++){
		indx = Index(*parX,0,2,i,0);
		om_ss[i]=v.d11[indx];
	}
	chebft_Extremes(om_ss,(*parX).ns[0],0);
	chder(0., 1., om_ss, om_sss,  (*parX).ns[0]);
	(*parX).J_infty     = pow(r0,3)*chebev(0.,1.,om_sss,(*parX).ns[0],0.)/12.;
	M_and_J         (*parX, v, w, G, 0, &((*parX).M), &((*parX).J));
//	Volume_integrals(*parX, v, w, G, &((*parX).M0), &((*parX).M_vol), &((*parX).J_vol), &((*parX).VI2), &((*parX).VI3));
	(*parX).h_max = h_max(*parX, v, w, G);
	(*parX).Rcirc_i = Rcirc(*parX, v, w, G, 0);
	(*parX).Rcirc_o = Rcirc(*parX, v, w, G, (*parX).ns[NDOM-1]-1);
}
//  -----------------------------------------------
void GetPhysicalParameters_Lin(parameters *parX, parameters *DparX,
       derivs_2D v,   derivs_1D w,  derivs_1D G,  derivs_2D V,
       derivs_2D Dv, derivs_1D Dw, derivs_1D DG, derivs_2D DV)
{	// parameters 1-6 already determined in  <Get_vwGpar_From_X>
	
	double ro = sqrt((*parX).xo),    ri  = sqrt((*parX).xi),
		  Dro = 0.5*(*DparX).xo/ro,  Dri = 0.5*(*DparX).xi/ri,
			x0 = X0FACT*(*parX).xo, Dx0 = X0FACT*(*DparX).xo;
	
	(*DparX).r_ratio    = (*parX).r_ratio*(Dri/ri - Dro/ro);
	(*DparX).beta_i     = D_ShedParameter(*parX, *DparX, 0, G, DG);
	(*DparX).beta_o     = D_ShedParameter(*parX, *DparX, (*parX).ns[NDOM-1]-1, G, DG);
//	(*DparX).M        = -sqrt((*parX).x0)*Dv.d1[Index(*parX, 0, 0, 0, 0)]; ?? this is original
	(*DparX).M_infty    = -sqrt(x0)*Dv.d1[Index(*parX, 0, 0, 0, 0)]  
	                      -0.5*Dx0/sqrt(x0)*v.d1[Index(*parX, 0, 0, 0, 0)];
}
//  -----------------------------------------------
void NonLinParamEqns(parameters par, parameters parX, derivs_2D v, derivs_1D w, derivs_1D G, derivs_2D V, double *FF)
{
	int n_v_2D=par.n_v_2D, ipar;

	GetPhysicalParameters_NonLin(&parX, v, w, G);
	
	for(ipar=0; ipar<NPAR; ipar++){
		switch(par.choice[ipar]){
			case CHOICE_V0        : FF[n_v_2D + ipar] = parX.V0      - par.V0;      break;
			case CHOICE_Om        : FF[n_v_2D + ipar] = parX.Om      - par.Om ;     break;
			case CHOICE_xi        : FF[n_v_2D + ipar] = parX.xi      - par.xi;      break;
			case CHOICE_xo        : FF[n_v_2D + ipar] = parX.xo      - par.xo;      break;
			case CHOICE_r_ratio   : FF[n_v_2D + ipar] = parX.r_ratio - par.r_ratio; break;
			case CHOICE_beta_i    : FF[n_v_2D + ipar] = parX.beta_i  - par.beta_i;  break;
			case CHOICE_beta_o    : FF[n_v_2D + ipar] = parX.beta_o  - par.beta_o;  break;
			case CHOICE_M         : FF[n_v_2D + ipar] = parX.M       - par.M;       break;
			case CHOICE_J         : FF[n_v_2D + ipar] = parX.J       - par.J;       break;
		}
	}
}
//  -----------------------------------------------
void LinParamEqns(parameters par, parameters parX, parameters DparX, 
					derivs_2D v, derivs_1D w, derivs_1D G, derivs_2D V, 
					derivs_2D Dv, derivs_1D Dw, derivs_1D DG, derivs_2D DV, double *JJDX)
{
	int n_v_2D=par.n_v_2D, ipar;

	GetPhysicalParameters_Lin(&parX, &DparX, v, w, G, V, Dv, Dw, DG, DV);

	for(ipar=0; ipar<NPAR; ipar++){
		switch(par.choice[ipar]){
			case CHOICE_V0        : JJDX[n_v_2D + ipar] = DparX.V0;      break;
			case CHOICE_Om        : JJDX[n_v_2D + ipar] = DparX.Om;      break;
			case CHOICE_xi        : JJDX[n_v_2D + ipar] = DparX.xi;      break;
			case CHOICE_xo        : JJDX[n_v_2D + ipar] = DparX.xo;      break;
			case CHOICE_r_ratio   : JJDX[n_v_2D + ipar] = DparX.r_ratio; break;
			case CHOICE_beta_i    : JJDX[n_v_2D + ipar] = DparX.beta_i;  break;
			case CHOICE_beta_o    : JJDX[n_v_2D + ipar] = DparX.beta_o;  break;
			case CHOICE_M         : JJDX[n_v_2D + ipar] = DparX.M;       break;
			case CHOICE_J         : JJDX[n_v_2D + ipar] = DparX.J;       break;
		}
	}
}
//  -----------------------------------------------
