#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include <time.h>
#include "utilities.h"
#include "Ring.h"

void surf_file(parameters par, derivs_1D G)
{
	double s, t=0.;
	int j, n = 50, idom = NDOM-1;
	coordinate xx, yy;
	derivs_1D Gcheb;
	FILE *stream;

	allocate_derivs_1D(&Gcheb, par.ns[NDOM-1]);

	for (j=0; j<par.ns[NDOM-1]; j++) Gcheb.d0[j] = G.d0[j];
	chebft_Extremes(Gcheb.d0, par.ns[NDOM-1], 0); // create Chebyshev coefficients for G
	chder(0., 1., Gcheb.d0, Gcheb.d1,  par.ns[NDOM-1]);
	chder(0., 1., Gcheb.d1, Gcheb.d11,  par.ns[NDOM-1]);

	stream = fopen("surface.dat", "w");

	for(j = 0; j<=n; j++){
		s = 1.*j/n;
		coord_3_all_st(par, s, t, Gcheb, &xx, &yy);
		fprintf(stream, "%e \t %e \n", xx.d0, yy.d0);
	}
	fclose(stream);
	free_derivs_1D(&Gcheb, par.ns[NDOM-1]);
}
// -----------------------------------------------------
void ergosphere_file(parameters par, derivs_2D v, derivs_1D w, derivs_1D G)
{
	int n=50, ipot, idom, i, j, ns, nt, Ns, Nt, n_theta, counter;
	double tolerance = 1.e-8, Laplace[NPOT], grad_grad[NPOT*NPOT], ss, tt, s, s0, s1, t, x, a, sa, y, b, sb,
		u, B, om, e2U, ux, Bx, omx, e2Ux, x_origin, r0, r1, r, theta,
		cheb_u[NMAX], cheb_B[NMAX], cheb_om[NMAX],
		cheb_ut[NMAX], cheb_Bt[NMAX], cheb_omt[NMAX],
		cheb_ux[NMAX], cheb_Bx[NMAX], cheb_omx[NMAX],***pot[NPOT-1];
	coordinate xx, yy;
	FILE *stream;
	derivs_1D Gcheb;
	derivs_2D V;
	st st_pair;


	idom = NDOM-1;
	ns = par.ns[idom]; Ns=ns-1;
	nt = par.nt[idom]; Nt=nt-1;

	allocate_derivs_2D(&V, NPOT);
	allocate_derivs_1D(&Gcheb, ns);

	for (i=0; i<ns; i++) Gcheb.d0[i] = G.d0[i];
	chebft_Extremes(Gcheb.d0, ns, 0); // create Chebyshev coefficients for G
	chder(0., 1., Gcheb.d0, Gcheb.d1,  ns);
	chder(0., 1., Gcheb.d1, Gcheb.d11,  ns);


// the position of the minimum value of e2U is determined in the ring for y=0
	t=1.; j=nt-1;
	for(i=0; i<ns; i++){ // here I create Chebyshev coefficients
		a=Pih*i/(ns-1), sa=sin(a), s=sa*sa;
		coord_st_To_xy(par, idom, i, j, s, t, G, &xx, &yy);
		NonLinPotentials(par, idom, i, j, s, t, xx, yy, v, w, G, V, Laplace, grad_grad);
		cheb_u[i]=V.d0[0]; cheb_B[i]=V.d0[1]; cheb_om[i]=V.d0[2];
		cheb_ux[i]=V.d1[0]; cheb_Bx[i]=V.d1[1]; cheb_omx[i]=V.d1[2];
	}
	chebft_Extremes(cheb_u, ns, 0); chebft_Extremes(cheb_ux, ns, 0);
	chebft_Extremes(cheb_B, ns, 0); chebft_Extremes(cheb_Bx, ns, 0);
	chebft_Extremes(cheb_om, ns, 0); chebft_Extremes(cheb_omx, ns, 0);
	s0=0.; s1=1.0; s=0.5*(s0+s1);
	while (fabs(s0-s1)>1.e-15){
		coord_3_all_st(par, s, t, Gcheb, &xx, &yy);
		x = xx.d0;
		u = chebev(0., 1., cheb_u, ns, s); ux = chebev(0., 1., cheb_ux, ns, s);
		B = chebev(0., 1., cheb_B, ns, s); Bx = chebev(0., 1., cheb_Bx, ns, s);
		om = chebev(0., 1., cheb_om, ns, s); omx = chebev(0., 1., cheb_omx, ns, s);
		e2Ux = 2.*B*exp(2.*u)*(Bx + B*ux) - exp(-2.*u)*om*om - 2.*x*om*exp(-2.*u)*(-ux*om + omx);
		if (e2Ux<0.) s0=s;
		else         s1=s;
		s = 0.5*(s0+s1);
	}
	e2U = B*B*exp(2.*u) - x*exp(-2.*u)*om*om;
	if (e2U>0.){ printf("Star possesses no ergoregion. Please perform `gnuplot surface.plt' \n\n");
		stream = fopen("ergosphere.dat", "w"); // empty file written
		fclose(stream);
	}
	else{
		printf("Star possesses an ergoregion. Please perform `gnuplot surface_and_ergo.plt' \n\n");
		stream = fopen("ergosphere.dat", "w");
		x_origin = x;
		for(ipot=0; ipot<NPOT-1; ipot++) pot[ipot]=d3tensor(0,NDOM-1,0,ns,0,nt);
		for(idom=0; idom<NDOM; idom++){
			for(i=0; i<ns; i++){
				a=Pih*i/Ns, sa=sin(a), s=sa*sa;
				for(j=0; j<nt; j++){
					b=Pih*j/Nt, sb=sin(b), t=sb*sb;
					coord_st_To_xy(par, idom, i, j, s, t, G, &xx, &yy);
					NonLinPotentials(par, idom, i, j, s, t, xx, yy, v, w, G, V, Laplace, grad_grad);
					for(ipot=0; ipot<NPOT-1; ipot++) pot[ipot][idom][i][j]=V.d0[ipot];
				}
			}
			for(ipot=0; ipot<NPOT-1; ipot++) chebft_Extremes_2D(pot[ipot][idom], ns, nt);
		}

		for (n_theta=0; n_theta<=n; n_theta++){
			theta = Pi*n_theta/n;
			r0 = 0.; r1 = sqrt(x_origin); r = 0.5*(r0+r1);
			e2U= 1.;
			counter = 0;
			while (fabs(e2U) > tolerance && counter < 50){// for y=0, x is increased outward until e2U changes its sign
				counter += 1;
				x = pow(sqrt(x_origin) - r*cos(theta),2);
				y = pow(r*sin(theta),2);
				st_pair = coord_xy_To_st_old(par, G, x, y);
				idom = st_pair.idom; s = st_pair.s; t = st_pair.t;
				u = chebevxy(0., 1., 0., 1., pot[0][idom], ns, nt, s, t);
				B = chebevxy(0., 1., 0., 1., pot[1][idom], ns, nt, s, t);
				om = chebevxy(0., 1., 0., 1., pot[2][idom], ns, nt, s, t);
				e2U = B*B*exp(2.*u) - x*exp(-2.*u)*om*om;
				if (e2U < 0.)       r0 = r;
				else                r1 = r;
				r = 0.5*(r0+r1);
			}
			if (fabs(e2U) > tolerance) printf("ergosphere_file failed to solve e2U=0 for theta=%e\n",theta);
//			if (fabs(e2U) > tolerance) printf("ergosphere_file failed to solve e2U=0 for theta=%e (e2U=%e, delr=%e counter=%d)\n",theta,e2U,r0-r1,counter);
			else fprintf(stream, "%e \t %e \t %e \t %d\n", x, y, e2U, idom);
			printf("finding ergosphere: %6.1lf%% done.\r", 100.*n_theta/n); 
			fflush(0);
		}
		for(ipot=0; ipot<NPOT-1; ipot++) free_d3tensor(pot[ipot],0,NDOM-1,0,ns,0,nt);
	}


	free_derivs_2D(&V, NPOT);
	free_derivs_1D(&Gcheb, par.ns[NDOM-1]);

}
// -----------------------------------------------------
void ergosphere_file_old(parameters par, derivs_2D v, derivs_1D w, derivs_1D G)
{
	int n=50, idom, i, j, ns, nt, n_theta, counter;
	double tolerance = 1.e-10, Laplace[NPOT], grad_grad[NPOT*NPOT], ss, tt, s, s0, s1, t, x, a, sa, y, b, sb,
		u, B, om, e2U, ux, Bx, omx, e2Ux, x_origin, r0, r1, r, theta,
		cheb_u[NMAX], cheb_B[NMAX], cheb_om[NMAX],
		cheb_ut[NMAX], cheb_Bt[NMAX], cheb_omt[NMAX],
		cheb_ux[NMAX], cheb_Bx[NMAX], cheb_omx[NMAX];
	coordinate xx, yy;
	FILE *stream;
	derivs_2D V;
	st st_pair;

	allocate_derivs_2D(&V, NPOT);


// the position of the minimum value of e2U is determined in the ring for y=0
	idom = NDOM-1;
	ns = par.ns[idom];
	nt = par.nt[idom];
	t=1.; j=nt-1;
	for(i=0; i<ns; i++){ // here I create Chebyshev coefficients
		a=Pih*i/(ns-1), sa=sin(a), s=sa*sa;
		coord_st_To_xy(par, idom, i, j, s, t, G, &xx, &yy);
		NonLinPotentials(par, idom, i, j, s, t, xx, yy, v, w, G, V, Laplace, grad_grad);
		cheb_u[i]=V.d0[0]; cheb_B[i]=V.d0[1]; cheb_om[i]=V.d0[2];
		cheb_ux[i]=V.d1[0]; cheb_Bx[i]=V.d1[1]; cheb_omx[i]=V.d1[2];
	}
	chebft_Extremes(cheb_u, ns, 0); chebft_Extremes(cheb_ux, ns, 0);
	chebft_Extremes(cheb_B, ns, 0); chebft_Extremes(cheb_Bx, ns, 0);
	chebft_Extremes(cheb_om, ns, 0); chebft_Extremes(cheb_omx, ns, 0);
	s0=0.; s1=1.0; s=0.5*(s0+s1);
	while (fabs(s0-s1)>1.e-15){
		coord_3_all_st_old(par, s, t, G, &xx, &yy);
		x = xx.d0;
		u = chebev(0., 1., cheb_u, ns, s); ux = chebev(0., 1., cheb_ux, ns, s);
		B = chebev(0., 1., cheb_B, ns, s); Bx = chebev(0., 1., cheb_Bx, ns, s);
		om = chebev(0., 1., cheb_om, ns, s); omx = chebev(0., 1., cheb_omx, ns, s);
		e2Ux = 2.*B*exp(2.*u)*(Bx + B*ux) - exp(-2.*u)*om*om - 2.*x*om*exp(-2.*u)*(-ux*om + omx);
		if (e2Ux<0.) s0=s;
		else         s1=s;
		s = 0.5*(s0+s1);
	}
	e2U = B*B*exp(2.*u) - x*exp(-2.*u)*om*om;
	if (e2U>0.){ printf("Star possesses no ergoregion. Please perform `gnuplot surface.plt' \n\n");
		stream = fopen("ergosphere.dat", "w"); // empty file written
		fclose(stream);
	}
	else{ 
		printf("Star possesses an ergoregion. Please perform `gnuplot surface_and_ergo.plt' \n\n");
		stream = fopen("ergosphere.dat", "w");
		x_origin = x;
		for (n_theta=0; n_theta<=n; n_theta++){
			theta = Pi*n_theta/n;
			r0 = 0.; r1 = sqrt(x_origin); r = 0.5*(r0+r1);
			e2U= 1.;
			counter = 0;
			while (fabs(e2U) > tolerance && counter < 100){// for y=0, x is increased outward until e2U changes its sign
				counter += 1;
				x = pow(sqrt(x_origin) - r*cos(theta),2);
				y = pow(r*sin(theta),2);
				st_pair = coord_xy_To_st_old(par, G, x, y);
				idom = st_pair.idom; s = st_pair.s; t = st_pair.t; ns = par.ns[idom]; nt = par.nt[idom];
				for(i=0; i<ns; i++){ // here I create Chebyshev coefficients
					a=Pih*i/(ns-1), sa=sin(a), ss=sa*sa;
					for(j=0; j<nt; j++){ // here I create Chebyshev coefficients
						b=Pih*i/(nt-1), sb=sin(b), tt=sb*sb;
						coord_st_To_xy(par, idom, i, j, ss, tt, G, &xx, &yy);
						NonLinPotentials(par, idom, i, j, ss, tt, xx, yy, v, w, G, V, Laplace, grad_grad);
						cheb_ut[j]=V.d0[0]; cheb_Bt[j]=V.d0[1]; cheb_omt[j]=V.d0[2];
					}
					chebft_Extremes(cheb_ut, nt, 0); chebft_Extremes(cheb_Bt, nt, 0); chebft_Extremes(cheb_omt, nt, 0);
					cheb_u[i] = chebev(0., 1., cheb_ut, nt, t);
					cheb_B[i] = chebev(0., 1., cheb_Bt, nt, t);
					cheb_om[i] = chebev(0., 1., cheb_omt, nt, t);
				}
				chebft_Extremes(cheb_u, ns, 0); chebft_Extremes(cheb_B, ns, 0); chebft_Extremes(cheb_om, ns, 0);
				u = chebev(0., 1., cheb_u, ns, s);
				B = chebev(0., 1., cheb_B, ns, s);
				om = chebev(0., 1., cheb_om, ns, s);
				e2U = B*B*exp(2.*u) - x*exp(-2.*u)*om*om;
				if (e2U < 0.)       r0 = r;
				else                r1 = r;
				r = 0.5*(r0+r1);
			}
			if (fabs(e2U) > tolerance) printf("ergosphere_file failed to solve e2U=0 for theta=%e\n",theta);
			else fprintf(stream, "%e \t %e \t %e \t %d\n", x, y, e2U, idom);
			printf("finding ergosphere: %6.1lf%% done.\r", 100.*n_theta/n); 
			fflush(0);
		}
	}

	free_derivs_2D(&V, NPOT);
}
